/*
 * tcpspy.c - watch incoming and outgoing connections 
 *
 * Copyright (c) 2000 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: tcpspy.c,v 1.7 2000/10/29 22:55:01 fyre Stab $
 */

#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>

static char *rcsid = "$Id: tcpspy.c,v 1.7 2000/10/29 22:55:01 fyre Stab $";

/*
 * Compile-time settings you may like to change.
 *
 * CONNTABLE_BUCKETS: The number of `buckets' in the connection hashtable,
 *                    this should be a positive prime number not much greater
 *                    than 65536. The default value of 101 is a suitable
 *                    tradeoff between execution speed and memory usage. If
 *                    your system has enough memory use 65537 here.
 *
 * MAX_WATCH_USERS:
 * MAX_IGNORE_USERS:  The maximum number of users that may be watched or
 *                    ignored. The defaults may not be sufficient if your
 *                    system has many users.
 *
 */
#ifndef CONNTABLE_BUCKETS
#define CONNTABLE_BUCKETS 101
#endif
#ifndef MAX_WATCH_USERS
#define MAX_WATCH_USERS 64
#endif
#ifndef MAX_IGNORE_USERS
#define MAX_IGNORE_USERS 64
#endif

uid_t watch_users[MAX_WATCH_USERS];
uid_t ignore_users[MAX_IGNORE_USERS];

static int stopped = 0;

/*
 * This structure holds everything we need to know about a connection. We
 * use unsigned long instead of (for example) uid_t, ino_t to make hashing
 * easier.
 */
typedef struct conn {
	unsigned long lcl;
	unsigned long lclp;
	unsigned long rmt;
	unsigned long rmtp;
	unsigned long uid;
	unsigned long ino;
	
	struct conn *next;
} conn_t;

typedef struct conntable {
	conn_t *buckets[CONNTABLE_BUCKETS];
} conntable_t;

int ct_init (conntable_t *ct);
int ct_hash (conn_t *c);
int ct_add (conntable_t *ct, conn_t *c);
int ct_find (conntable_t *ct, conn_t *c);
int ct_read (conntable_t *ct);
int ct_free (conntable_t *ct);
void logconn (conn_t *c, const char *action);
void compare (conntable_t *old, conntable_t *new);
void panic (const char *fmt, ...);
void stopsig (int signal);

/*
 * Stop the program when certain signals are received. 
 */
void stopsig (int signal)
{
	syslog (LOG_NOTICE, "shutting down on signal");
	stopped = 1;
}

/*
 * Scream to syslog and exit.
 */
void panic (const char *fmt, ...)
{
	va_list ap;

	va_start (ap, fmt);
	vsyslog (LOG_ERR, fmt, ap);
	va_end (ap);

	exit (EXIT_FAILURE);
}

/*
 * Initialise a connection table (hashtable).
 */
int ct_init (conntable_t *ct)
{
	int i;
	
	assert (ct != NULL);

	for (i = 0; i < CONNTABLE_BUCKETS; i++)
		ct->buckets[i] = NULL;

	return 1;
}

/*
 * Simple hash function for connections. There is definitely room for
 * improvement here.
 */
int ct_hash (conn_t *c)
{
	unsigned long h;

	assert (c != NULL);

	h = c->lcl ^ c->lclp ^ c->rmt ^ c->rmtp ^ c->uid ^ c->ino;

	return h % CONNTABLE_BUCKETS;
}

/*
 * Add a connection to the connection table.
 */
int ct_add (conntable_t *ct, conn_t *c)
{
	conn_t *old, *newc;
	int bi;
	
	assert (ct != NULL);
	assert (c != NULL);

	newc = (conn_t *) malloc (sizeof (conn_t));
	if (newc == NULL) panic ("out of memory in ct_add");	

	memcpy (newc, c, sizeof (conn_t));

	bi = ct_hash (c);
	old = ct->buckets[bi];
	ct->buckets[bi] = newc;
	ct->buckets[bi]->next = old;

	return 1;
}

/*
 * Find a connection in a table, return nonzero if found, zero otherwise.
 */
int ct_find (conntable_t *ct, conn_t *c)
{
	conn_t *bucket;
	
	assert (ct != NULL);
	assert (c != NULL);

	bucket = ct->buckets[ct_hash (c)];
	while (bucket != NULL) {
		if ((c->lcl == bucket->lcl) && (c->lclp == bucket->lclp) && 
			(c->rmt == bucket->rmt) && (c->rmtp == bucket->rmtp) &&
			(c->uid == bucket->uid) && (c->ino == bucket->ino)) {
			return 1;
		}
		bucket = bucket->next;
	}

	return 0;
}

/*
 * Read /proc/net/tcp and add all connections to the table.
 */
int ct_read (conntable_t *ct)
{
	FILE *fp;
	char buf[1024];
	conn_t c;
	unsigned long st;
	int i, watched = 1;
	
	assert (ct != NULL);

	fp = fopen ("/proc/net/tcp", "r");
	if (fp == NULL) panic ("/proc/net/tcp: %s", strerror (errno));

	if (fgets (buf, sizeof (buf), fp) == NULL) {
		panic ("/proc/net/tcp: missing header");
		return 0;
	}
	
	while (fgets (buf, sizeof (buf), fp) != NULL) {
		if (sscanf (buf, "%*d: %lx:%lx %lx:%lx %lx %*x:%*x %*x:%*x %*x %lu %*d %lu", &c.lcl, &c.lclp, &c.rmt, &c.rmtp, &st, &c.uid, &c.ino) != 7) {
			panic ("/proc/net/tcp: bad file format");
			exit (EXIT_FAILURE);
		}
		if ((c.ino == 0) || (st != 0x01)) continue;

		if (watch_users[0] != (uid_t) -1) {
			watched = 0;
			for (i = 0; watch_users[i] != (uid_t) -1; i++)
				if (watch_users[i] == c.uid) watched = 1;
		} else watched = 1;
		if (ignore_users[0] != (uid_t) -1) {
			for (i = 0; ignore_users[i] != (uid_t) -1; i++)
				if (ignore_users[i] == c.uid) watched = 0;
		}

		if (watched == 0) continue;

		if (ct_add (ct, &c) == 0) {
			fclose (fp);
			return 0;
		}
	}

	fclose (fp);

	return 1;
}

/*
 * Free a connection table.
 */
int ct_free (conntable_t *ct)
{
	int i;

	assert (ct != NULL);

	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *c0, *c1;

		c0 = ct->buckets[i];
		while (c0 != NULL) {
			c1 = c0->next;
			free (c0);
			c0 = c1;
		}
		ct->buckets[i] = NULL;
	}

	return 1;
}

/*
 * Log a connection/disconnected.
 */
void logconn (conn_t *c, const char *action)
{
	struct in_addr in;
	char laddr[256], raddr[256];
	char *n;
	struct passwd *pw;
	char uidbuf[256];
	char *user;

	assert (c != NULL);
	assert (action != NULL);

	in.s_addr = c->lcl;
	n = inet_ntoa (in);
	strcpy (laddr, n);
	in.s_addr = c->rmt;
	n = inet_ntoa (in);
	strcpy (raddr, n);

	if ((pw = getpwuid ((uid_t) c->uid)) == NULL) {
		sprintf (uidbuf, "<uid %u>", (unsigned int)(uid_t)c->uid);
		user = uidbuf;
	} else {
		user = pw->pw_name;
	}

	syslog (LOG_NOTICE, "%s: user %s, local %s:%u, remote %s:%u",
			action, user, laddr, c->lclp, raddr, c->rmtp);
}

/*
 * Compare the `old' and `new' tables, logging any differences.
 */
void compare (conntable_t *old, conntable_t *new)
{
	int i;
	
	assert (old != NULL);
	assert (new != NULL);
	
	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *bucket;

		bucket = old->buckets[i];
		while (bucket != NULL) {
			if (ct_find (new, bucket) == 0)
				logconn (bucket, "disconnect");	
			bucket = bucket->next;
		}
	}

	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *bucket;

		bucket = new->buckets[i];
		while (bucket != NULL) {
			if (ct_find (old, bucket) == 0)
				logconn (bucket, "connect");
			bucket = bucket->next;
		}
	}
}

uid_t lookup_user (const char *u)
{
	assert (u != NULL);

	if (isdigit (*u)) return atoi (u);
	else {
		struct passwd *pw;
		if ((pw = getpwnam (u)) == NULL) {
			fprintf (stderr, "tcpspy: user `%s' not found\n", u);
			exit (EXIT_FAILURE);
		}
		else return pw->pw_uid;
	}
}

void usage (void)
{
	fprintf (stderr, "usage: tcpspy [-u username] [interval]\n");
	exit (EXIT_FAILURE);
}

int main (int argc, char *argv[])
{
	conntable_t old, new;
	unsigned long interval;
	pid_t p;
	int wu = 0, iu = 0;
	int ch;
	uid_t dropuser = -1;
	gid_t dropgroup = -1;
	int debug = 0;

	/*
	 * Parse our arguments.
	 */
	opterr = 0;
	while ((ch = getopt (argc, argv, "dcw:i:u:U:G:")) != -1) {
		switch (ch) {
			case 'u':
				fprintf (stderr, "tcpspy: warning: -u option is deprecated\n");
				/* fall through */
			case 'w':
				if (wu >= MAX_WATCH_USERS - 1) {
					fprintf (stderr, "tcpspy: too many watched users\n");
					exit (EXIT_FAILURE);
				}
				watch_users[wu++] = lookup_user (optarg);
				break;
			case 'i':
				if (iu >= MAX_IGNORE_USERS - 1) {
					fprintf (stderr, "tcpspy: too many ignored users\n");
					exit (EXIT_FAILURE);
				}
				ignore_users[iu++] = lookup_user (optarg);
				break;
			case 'U':
				if (isdigit (*optarg))
					dropuser = atoi (optarg);
				else {
					struct passwd *pw;
					if ((pw = getpwnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: user `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropuser = pw->pw_uid;
				}
				break;
			case 'G':
				if (isdigit (*optarg))
					dropgroup = atoi (optarg);
				else {
					struct group *gr;
					if ((gr = getgrnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: group `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropgroup = gr->gr_gid;
				}
				break;
			case 'd':
				debug = 1;
				break;
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (*argv != NULL) {
		interval = atoi (*argv);
		if (interval == 0) {
			fprintf (stderr, "tcpspy: bad interval\n");
			exit (EXIT_FAILURE);
		}
	} else interval = 1000;

	watch_users[wu] = (uid_t) -1;
	ignore_users[iu] = (uid_t) -1;

	/*
	 * Become an unprivileged user for safety purposes if requested.
	 */
	if (dropgroup != (gid_t) -1) {
		printf ("setgid %d\n", dropgroup);
		if (setgid (dropgroup) < 0) {
			fprintf (stderr, "tcpspy: setgid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}
	if (dropuser != (uid_t) -1) {
		if (setuid (dropuser) < 0) {
			fprintf (stderr, "tcpspy: setuid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}

	/*
	 * Become a daemon..
	 */

	if (debug == 0) {
		p = fork();
		if (p < 0) panic ("fork failed: %s", strerror (errno));
		else if (p != 0) {
			printf ("tcpspy 1.1 started (pid %d)\n", (int) p);
			exit (EXIT_SUCCESS);
		}

		ioctl (STDIN_FILENO, TIOCNOTTY, NULL);
		close (STDIN_FILENO); 
		close (STDOUT_FILENO); 
		close (STDERR_FILENO);
	} else {
		p = getpid();
		printf ("tcpspy 1.1 started (pid %d) (debug)\n", p);
	}

	signal (SIGTERM, stopsig);
	if (debug != 0) signal (SIGINT, stopsig);

	/*
	 * Initialisation's done, let's spy on the evil lusers ;)
	 */

	openlog ("tcpspy", LOG_PID, FACILITY);
	syslog (LOG_NOTICE, "tcpspy started, monitoring connections");

	if (ct_init (&old) == 0) panic ("ct_init failed");
	if (ct_read (&old) == 0) panic ("ct_read failed");
	
	while (stopped == 0) {
		usleep (interval * 1000);
		if (ct_init (&new) == 0) panic ("ct_init failed");
		if (ct_read (&new) == 0) panic ("ct_read failed");
		compare (&old, &new);
		if (ct_free (&old) == 0) panic ("ct_free failed");
		memcpy (&old, &new, sizeof (conntable_t));
	}

	if (ct_free (&old) == 0) panic ("ct_free failed");
	closelog ();

	return EXIT_SUCCESS;
}
